@@ -35,8 +35,13 @@ struct ChainNameGenerator {
   }
 
   std::string make_short_name(const std::string& preferred) {
-    static const char symbols[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-                                  "abcdefghijklmnopqrstuvwxyz0123456789";
+    static const char symbols[] = {
+      'A','B','C','D','E','F','G','H','I','J','K','L','M',
+      'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
+      'a','b','c','d','e','f','g','h','i','j','k','l','m',
+      'n','o','p','q','r','s','t','u','v','w','x','y','z',
+      '0','1','2','3','4','5','6','7','8','9'
+    };
     if (!has(preferred))
       return added(preferred);
     std::string name(1, 'A');
@@ -226,23 +231,34 @@ inline void shorten_chain_names(Structure& st) {
 }
 
 
-inline void expand_ncs(gemmi::Structure& st, HowToNameCopiedChain how) {
-  for (gemmi::Model& model : st.models) {
+inline void expand_ncs(Structure& st, HowToNameCopiedChain how) {
+  size_t orig_conn_size = st.connections.size();
+  for (Model& model : st.models) {
+    if (how == HowToNameCopiedChain::Dup) {
+      // change segment of original chains to "0" - is this a good idea?
+      for (Chain& chain : model.chains)
+        for (Residue& res : chain.residues)
+          res.segment = "0";
+    }
     size_t orig_size = model.chains.size();
-    gemmi::ChainNameGenerator namegen(model, how);
-    for (const gemmi::NcsOp& op : st.ncs)
+    ChainNameGenerator namegen(model, how);
+    for (const NcsOp& op : st.ncs)
       if (!op.given) {
+        std::map<std::string, std::string> chain_mapping;
         for (size_t i = 0; i != orig_size; ++i) {
-          if (how == HowToNameCopiedChain::Dup)
-            for (gemmi::Residue& res : model.chains[i].residues)
-              res.segment = "0";
-
           model.chains.push_back(model.chains[i]);
-          gemmi::Chain& new_chain = model.chains.back();
-          new_chain.name = namegen.make_new_name(new_chain.name, (int)i+1);
+          Chain& new_chain = model.chains.back();
+          const std::string& old_name = model.chains[i].name;
+          auto it = chain_mapping.find(old_name);
+          if (it == chain_mapping.end()) {
+            new_chain.name = namegen.make_new_name(old_name, (int)i+1);
+            chain_mapping.emplace(old_name, new_chain.name);
+          } else {
+            new_chain.name = it->second;
+          }
 
-          for (gemmi::Residue& res : new_chain.residues) {
-            for (gemmi::Atom& a : res.atoms)
+          for (Residue& res : new_chain.residues) {
+            for (Atom& a : res.atoms)
               transform_atom(a, op.tr);
             if (!res.subchain.empty())
               res.subchain = new_chain.name + ":" + res.subchain;
@@ -250,13 +266,83 @@ inline void expand_ncs(gemmi::Structure& st, HowToNameCopiedChain how) {
               res.segment = op.id;
           }
         }
+        // add connections when processing the first model
+        if (&model == &st.models[0]) {
+          for (size_t i = 0; i != orig_conn_size; ++i) {
+            st.connections.push_back(st.connections[i]);
+            Connection& c = st.connections.back();
+            c.name += '-';
+            c.name += op.id;
+            for (int j = 0; j < 2; ++j) {
+              AtomAddress& aa = j == 0 ? c.partner1 : c.partner2;
+              if (how == HowToNameCopiedChain::Dup) {
+                aa.res_id.segment = op.id;
+              } else {
+                auto it = chain_mapping.find(aa.chain_name);
+                if (it != chain_mapping.end())
+                  aa.chain_name = it->second;
+                else
+                  st.connections.pop_back();
+              }
+            }
+          }
+        }
       }
   }
-  for (gemmi::NcsOp& op : st.ncs)
+  // adjust connections after changing segment of original chains to "0"
+  if (how == HowToNameCopiedChain::Dup) {
+    for (size_t i = 0; i != orig_conn_size; ++i) {
+      st.connections[i].partner1.res_id.segment = "0";
+      st.connections[i].partner2.res_id.segment = "0";
+    }
+  }
+  for (NcsOp& op : st.ncs)
     op.given = true;
   st.setup_cell_images();
 }
 
+inline std::vector<Chain> split_chain_by_segments(Chain& orig, ChainNameGenerator& namegen) {
+  std::vector<Chain> chains;
+  std::vector<Residue> orig_res;
+  orig_res.swap(orig.residues);
+  int n = 0;
+  for (auto start = orig_res.begin(); start != orig_res.end(); ) {
+    const std::string& seg = start->segment;
+    auto ch = std::find_if(chains.begin(), chains.end(), [&](Chain& c) {
+                return !c.residues.empty() && c.residues[0].segment == seg; });
+    if (ch == chains.end()) {
+      chains.push_back(orig);
+      ch = chains.end() - 1;
+      // Naming. Here, Dup means chain name + segment.
+      switch (namegen.how) {
+        case HowToNameCopiedChain::Short:
+          ch->name = namegen.make_short_name(ch->name + seg);
+          break;
+        case HowToNameCopiedChain::AddNumber:
+          ch->name = namegen.make_name_with_numeric_postfix(ch->name, ++n);
+          break;
+        case HowToNameCopiedChain::Dup:
+          ch->name += seg;
+          break;
+      }
+    }
+    auto end = std::find_if(start, orig_res.end(),
+                            [&](Residue& r) { return r.segment != seg; });
+    ch->residues.insert(ch->residues.end(), std::make_move_iterator(start),
+                                            std::make_move_iterator(end));
+    start = end;
+  }
+  return chains;
+}
+
+// HowToNameCopiedChain::Dup adds segment name to
+inline void split_chains_by_segments(Model& model, HowToNameCopiedChain how) {
+  ChainNameGenerator namegen(how);
+  std::vector<Chain> new_chains;
+  for (Chain& chain : model.chains)
+    vector_move_extend(new_chains, split_chain_by_segments(chain, namegen));
+  model.chains = std::move(new_chains);
+}
 
 } // namespace gemmi
 #endif
